# -*- coding: utf-8 -*-

import time
import os
import subprocess
import sys
import signal
import copy
from collections import Counter
import zmq
import setproctitle

from .request import Request
from .connection import Connection
from .mixins import RoutesMixin, AppEventsMixin
from ..share.log import logger
from ..share import constants
from ..share.config import ConfigAttribute, Config
from ..share.utils import import_module_or_string


class Worker(RoutesMixin, AppEventsMixin):
    ############################## configurable begin ##############################

    # 显示的进程名
    name = ConfigAttribute('NAME')
    # 消息协议类
    box_class = ConfigAttribute('BOX_CLASS',
                                get_converter=import_module_or_string)

    # 调试模式
    debug = ConfigAttribute('DEBUG')

    # 启动多少个子进程
    spawn_count = ConfigAttribute('WORKER_SPAWN_COUNT')

    # 最多回应一次
    rsp_once = ConfigAttribute('WORKER_RSP_ONCE')
    # 处理task超时(秒). 超过后worker会自杀. None 代表永不超时
    work_timeout = ConfigAttribute('WORKER_WORK_TIMEOUT')
    # 停止子进程超时(秒). 使用 TERM 进行停止时，如果超时未停止会发送KILL信号
    stop_timeout = ConfigAttribute('WORKER_STOP_TIMEOUT')

    ############################## configurable end   ##############################
    # connection 类
    connection_class = Connection
    # request 类
    request_class = Request

    config = None

    got_first_request = False
    blueprints = None
    # 是否有效(父进程中代表程序有效，子进程中代表worker是否有效)
    enable = True
    # 子进程列表
    processes = None

    # 连接到forwarder的连接
    forwarder_client = None

    def __init__(self):
        RoutesMixin.__init__(self)
        AppEventsMixin.__init__(self)
        self.config = Config(defaults=constants.DEFAULT_CONFIG)
        self.blueprints = list()
        self.processes = list()

    def register_blueprint(self, blueprint):
        blueprint.register_to_app(self)

    def run(self, debug=None):
        """

        :param debug:
        :return:
        """
        self._validate_cmds()

        if debug is not None:
            self.debug = debug

        if os.getenv(constants.WORKER_ENV_KEY) != 'true':
            # 主进程
            logger.info('Connect to server , debug: %s, workers: %s',
                        self.debug, self.spawn_count)

            # 设置进程名
            setproctitle.setproctitle(self._make_proc_name('worker:master'))
            # 只能在主线程里面设置signals
            self._handle_parent_proc_signals()
            self._spawn_workers(self.spawn_count)
        else:
            # 子进程
            setproctitle.setproctitle(self._make_proc_name('worker:worker'))
            self._worker_run()

    def _make_proc_name(self, subtitle):
        """
        获取进程名称
        :param subtitle:
        :return:
        """
        proc_name = '[%s:%s %s] %s' % (
            constants.NAME,
            subtitle,
            self.name,
            ' '.join([sys.executable] + sys.argv)
        )

        return proc_name

    def _validate_cmds(self):
        """
        确保 cmd 没有重复
        :return:
        """

        cmd_list = list(self.rule_map.keys())

        for bp in self.blueprints:
            cmd_list.extend(bp.rule_map.keys())

        duplicate_cmds = (Counter(cmd_list) - Counter(set(cmd_list))).keys()

        assert not duplicate_cmds, 'duplicate cmds: %s' % duplicate_cmds

    def _on_worker_run(self):
        # 连接到结果server
        self._connect_to_forwarder_server()

        self.events.create_worker()
        for bp in self.blueprints:
            bp.events.create_app_worker()

    def _worker_run(self):
        self._handle_child_proc_signals()

        self._on_worker_run()

        try:
            self._serve_forever()
        except KeyboardInterrupt:
            pass
        except:
            logger.error('exc occur.', exc_info=True)

    def _spawn_workers(self, workers):
        worker_env = copy.deepcopy(os.environ)
        worker_env.update({
            constants.WORKER_ENV_KEY: 'true'
        })

        def start_worker_process():
            args = [sys.executable] + sys.argv
            try:
                return subprocess.Popen(args, env=worker_env)
            except:
                logger.error('exc occur. args: %s, env: %s', args, worker_env, exc_info=True)
                return None

        for it in xrange(0, workers):
            p = start_worker_process()
            self.processes.append(p)

        while 1:
            for idx, p in enumerate(self.processes):
                if p and p.poll() is not None:
                    # 说明退出了
                    self.processes[idx] = None

                    if self.enable:
                        # 如果还要继续服务
                        p = start_worker_process()
                        self.processes[idx] = p

            if not filter(lambda x: x, self.processes):
                # 没活着的了
                break

            # 时间短点，退出的快一些
            time.sleep(0.1)

    def _handle_parent_proc_signals(self):
        def exit_handler(signum, frame):
            self.enable = False

            # 如果是终端直接CTRL-C，子进程自然会在父进程之后收到INT信号，不需要再写代码发送
            # 如果直接kill -INT $parent_pid，子进程不会自动收到INT
            # 所以这里可能会导致重复发送的问题，重复发送会导致一些子进程异常，所以在子进程内部有做重复处理判断。
            for p in self.processes:
                if p:
                    p.send_signal(signum)

            # https://docs.python.org/2/library/signal.html#signal.alarm
            if self.stop_timeout is not None:
                signal.alarm(self.stop_timeout)

        def final_kill_handler(signum, frame):
            if not self.enable:
                # 只有满足了not enable，才发送term命令
                for p in self.processes:
                    if p:
                        p.send_signal(signal.SIGKILL)

        def safe_stop_handler(signum, frame):
            """
            等所有子进程结束，父进程也退出
            """
            self.enable = False

            for p in self.processes:
                if p:
                    p.send_signal(signal.SIGTERM)

            if self.stop_timeout is not None:
                signal.alarm(self.stop_timeout)

        def safe_reload_handler(signum, frame):
            """
            让所有子进程重新加载
            """
            for p in self.processes:
                if p:
                    p.send_signal(signal.SIGHUP)

        # INT, QUIT为强制结束
        signal.signal(signal.SIGINT, exit_handler)
        signal.signal(signal.SIGQUIT, exit_handler)
        # TERM为安全结束
        signal.signal(signal.SIGTERM, safe_stop_handler)
        # HUP为热更新
        signal.signal(signal.SIGHUP, safe_reload_handler)
        # 最终判决，KILL掉子进程
        signal.signal(signal.SIGALRM, final_kill_handler)

    def _handle_child_proc_signals(self):
        def exit_handler(signum, frame):
            # 防止重复处理KeyboardInterrupt，导致抛出异常
            if self.enable:
                self.enable = False
                raise KeyboardInterrupt

        def safe_stop_handler(signum, frame):
            self.enable = False

        # 强制结束，抛出异常终止程序进行
        signal.signal(signal.SIGINT, exit_handler)
        signal.signal(signal.SIGQUIT, exit_handler)
        # 安全停止
        signal.signal(signal.SIGTERM, safe_stop_handler)
        signal.signal(signal.SIGHUP, safe_stop_handler)

    def _serve_forever(self):
        conn = self.connection_class(self,
                                     self.config['GATEWAY_INNER_ADDRESS_LIST'],
                                     self.config['WORKER_CONN_TIMEOUT'])
        conn.run()

    def _connect_to_forwarder_server(self):
        """
        连接到结果服务器
        :return:
        """

        ctx = zmq.Context()
        self.forwarder_client = ctx.socket(zmq.PUSH)
        for address in self.config['FORWARDER_INPUT_ADDRESS_LIST']:
            self.forwarder_client.connect(address)

    def __repr__(self):
        return '<%s name: %s>' % (
            type(self).__name__, self.name
        )



# -*- coding: utf-8 -*-

from multiprocessing import Process
import time
from log import logger


class ProcMgr(object):
    enable = True
    processes = None

    def __init__(self):
        self.processes = []

    def spawn_workers(self, workers, target):
        def start_worker_process(index):
            try:
                inner_p = Process(target=target, args=(index,))
                # 当前进程daemon默认是False，改成True将启动不了子进程
                # 但是子进程要设置daemon为True，这样父进程退出，子进程会被强制关闭
                # 现在父进程会在子进程之后推出，没必要设置了
                # inner_p.daemon = True
                inner_p.start()
                return inner_p
            except:
                # 因为很有可能遇到文件句柄不够用的情况
                # 而不进行任何异常捕获的话，会导致master进程死掉
                logger.error('exc occur. target: %s, index: %s', target, index, exc_info=True)
                return None

        for it in xrange(0, workers):
            p = start_worker_process(it)
            self.processes.append(p)

        while 1:
            for idx, p in enumerate(self.processes):
                if p and not p.is_alive():
                    self.processes[idx] = None

                    if self.enable:
                        p = start_worker_process(idx)
                        self.processes[idx] = p

            if not filter(lambda x: x, self.processes):
                # 没活着的了
                break

            # 时间短点，退出的快一些
            time.sleep(0.1)

    def terminate_all(self):

        # 如果是终端直接CTRL-C，子进程自然会在父进程之后收到INT信号，不需要再写代码发送
        # 如果直接kill -INT $parent_pid，子进程不会自动收到INT
        # 所以这里可能会导致重复发送的问题，重复发送会导致一些子进程异常，所以在子进程内部有做重复处理判断。
        for p in self.processes:
            if p:
                p.terminate()



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: shine.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)




DESCRIPTOR = _descriptor.FileDescriptor(
  name='shine.proto',
  package='shine.proto',
  serialized_pb='\n\x0bshine.proto\x12\x0bshine.proto\"\x8c\x01\n\x04Task\x12\x0f\n\x07node_id\x18\x01 \x01(\x0c\x12\x11\n\tclient_id\x18\x02 \x01(\x0c\x12\x11\n\tclient_ip\x18\x03 \x01(\t\x12\r\n\x05inner\x18\x04 \x01(\x05\x12\x0b\n\x03\x63md\x18\x05 \x01(\x05\x12\x0e\n\x03uid\x18\x06 \x01(\x03:\x01\x30\x12\x13\n\x08userdata\x18\x07 \x01(\x03:\x01\x30\x12\x0c\n\x04\x62ody\x18\x08 \x01(\x0c\"|\n\nRspToUsers\x12)\n\x04rows\x18\x01 \x03(\x0b\x32\x1b.shine.proto.RspToUsers.Row\x1a\x43\n\x03Row\x12\x0c\n\x04uids\x18\x01 \x03(\x03\x12\x0b\n\x03\x62uf\x18\x02 \x01(\x0c\x12\x10\n\x08userdata\x18\x03 \x01(\x03\x12\x0f\n\x07\x65xclude\x18\x04 \x03(\x03\"=\n\nCloseUsers\x12\x0c\n\x04uids\x18\x01 \x03(\x03\x12\x10\n\x08userdata\x18\x02 \x01(\x03\x12\x0f\n\x07\x65xclude\x18\x03 \x03(\x03\x42\x1d\n\x14\x63n.vimer.shine.protoB\x05Proto')




_TASK = _descriptor.Descriptor(
  name='Task',
  full_name='shine.proto.Task',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='node_id', full_name='shine.proto.Task.node_id', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='client_id', full_name='shine.proto.Task.client_id', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='client_ip', full_name='shine.proto.Task.client_ip', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='inner', full_name='shine.proto.Task.inner', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='cmd', full_name='shine.proto.Task.cmd', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='uid', full_name='shine.proto.Task.uid', index=5,
      number=6, type=3, cpp_type=2, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='userdata', full_name='shine.proto.Task.userdata', index=6,
      number=7, type=3, cpp_type=2, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='body', full_name='shine.proto.Task.body', index=7,
      number=8, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=29,
  serialized_end=169,
)


_RSPTOUSERS_ROW = _descriptor.Descriptor(
  name='Row',
  full_name='shine.proto.RspToUsers.Row',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='uids', full_name='shine.proto.RspToUsers.Row.uids', index=0,
      number=1, type=3, cpp_type=2, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='buf', full_name='shine.proto.RspToUsers.Row.buf', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='userdata', full_name='shine.proto.RspToUsers.Row.userdata', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='exclude', full_name='shine.proto.RspToUsers.Row.exclude', index=3,
      number=4, type=3, cpp_type=2, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=228,
  serialized_end=295,
)

_RSPTOUSERS = _descriptor.Descriptor(
  name='RspToUsers',
  full_name='shine.proto.RspToUsers',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='rows', full_name='shine.proto.RspToUsers.rows', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_RSPTOUSERS_ROW, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=171,
  serialized_end=295,
)


_CLOSEUSERS = _descriptor.Descriptor(
  name='CloseUsers',
  full_name='shine.proto.CloseUsers',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='uids', full_name='shine.proto.CloseUsers.uids', index=0,
      number=1, type=3, cpp_type=2, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='userdata', full_name='shine.proto.CloseUsers.userdata', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='exclude', full_name='shine.proto.CloseUsers.exclude', index=2,
      number=3, type=3, cpp_type=2, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=297,
  serialized_end=358,
)

_RSPTOUSERS_ROW.containing_type = _RSPTOUSERS;
_RSPTOUSERS.fields_by_name['rows'].message_type = _RSPTOUSERS_ROW
DESCRIPTOR.message_types_by_name['Task'] = _TASK
DESCRIPTOR.message_types_by_name['RspToUsers'] = _RSPTOUSERS
DESCRIPTOR.message_types_by_name['CloseUsers'] = _CLOSEUSERS

class Task(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _TASK

  # @@protoc_insertion_point(class_scope:shine.proto.Task)

class RspToUsers(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType

  class Row(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _RSPTOUSERS_ROW

    # @@protoc_insertion_point(class_scope:shine.proto.RspToUsers.Row)
  DESCRIPTOR = _RSPTOUSERS

  # @@protoc_insertion_point(class_scope:shine.proto.RspToUsers)

class CloseUsers(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _CLOSEUSERS

  # @@protoc_insertion_point(class_scope:shine.proto.CloseUsers)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), '\n\024cn.vimer.shine.protoB\005Proto')
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-


import random
import signal
import sys
import setproctitle
import gevent
from gevent.queue import Queue
import zmq.green as zmq  # for gevent
from collections import defaultdict
from ..share.proc_mgr import ProcMgr
from ..share.log import logger
from ..share import constants, shine_pb2
from ..share.config import ConfigAttribute, Config
from ..share.share_store import ShareStore


class Forwarder(object):
    ############################## configurable begin ##############################

    name = ConfigAttribute('NAME')
    debug = ConfigAttribute('DEBUG')

    ############################## configurable end   ##############################

    config = None

    proc_mgr = None
    input_server = None
    output_server = None

    # 等待处理的队列[data, ]
    to_deal_queue = None
    # 等待发送的队列[(topic, data or task), ]
    to_send_queue = None

    share_store = None

    def __init__(self):
        self.config = Config(defaults=constants.DEFAULT_CONFIG)
        self.proc_mgr = ProcMgr()
        self.to_deal_queue = Queue()
        self.to_send_queue = Queue()

    def run(self, debug=None):
        """
        启动
        :param debug: 是否debug
        :return:
        """

        assert len(self.config['FORWARDER_INPUT_ADDRESS_LIST']) == len(self.config['FORWARDER_OUTPUT_ADDRESS_LIST'])

        if self.config['REDIS_URL']:
            import redis
            rds = redis.from_url(self.config['REDIS_URL'])
            self.share_store = ShareStore(rds,
                                          self.config['REDIS_KEY_SHARE_PREFIX'] + self.config['REDIS_USER_KEY_PREFIX'],
                                          self.config['REDIS_KEY_SHARE_PREFIX'] + self.config['REDIS_NODES_KEY'],
                                          self.config['REDIS_USER_MAX_AGE']
                                          )

        if debug is not None:
            self.debug = debug

        workers = len(self.config['FORWARDER_INPUT_ADDRESS_LIST'])

        def run_wrapper():
            logger.info('Running server, debug: %s, workers: %s',
                        self.debug, workers)

            setproctitle.setproctitle(self._make_proc_name('forwarder:master'))
            # 只能在主线程里面设置signals
            self._handle_parent_proc_signals()
            self.proc_mgr.spawn_workers(workers, self._worker_run)

        run_wrapper()

    def _make_proc_name(self, subtitle):
        """
        获取进程名称
        :param subtitle:
        :return:
        """
        proc_name = '[%s:%s %s] %s' % (
            constants.NAME,
            subtitle,
            self.name,
            ' '.join([sys.executable] + sys.argv)
        )

        return proc_name

    def _make_redis_key(self, uid):
        return self.config['REDIS_KEY_SHARE_PREFIX'] + self.config['REDIS_USER_KEY_PREFIX'] + str(uid)

    def _handle_task_forever(self):
        """
        一直在处理
        :return:
        """

        while 1:
            data = self.to_deal_queue.get()

            task = shine_pb2.Task()
            task.ParseFromString(data)

            logger.debug('task:\n%s', task)

            if task.cmd in (constants.CMD_WRITE_TO_CLIENT,
                            constants.CMD_LOGIN_CLIENT,
                            constants.CMD_LOGOUT_CLIENT,
                            constants.CMD_CLOSE_CLIENT):
                # 原样处理过去
                # 给data的好处是，就不用再序列化了
                self.to_send_queue.put((task.node_id, data))
            elif task.cmd == constants.CMD_WRITE_TO_WORKER:
                if task.HasField('node_id'):
                    self.to_send_queue.put((task.node_id, data))
                else:
                    # 随机选择一个node_id
                    node_id_list = self.share_store.get_nodes()
                    if node_id_list:
                        self.to_send_queue.put((random.choice(list(node_id_list)), data))
                    else:
                        logger.error('node_id_list is empty.')

            elif task.cmd == constants.CMD_WRITE_TO_USERS:
                if not self.share_store:
                    # 直接转发就好
                    self.to_send_queue.put((task.node_id, data))
                else:
                    # 这个是所有row合并在一起的uid_list，是为了快速获取node_id而用的
                    merged_uid_list = set()
                    rsp = shine_pb2.RspToUsers()
                    rsp.ParseFromString(task.body)
                    for row in rsp.rows:
                        merged_uid_list.update(set(row.uids))

                    merged_uid_list = list(merged_uid_list)  # 一定要变回来

                    node_id_to_uid_dict = self.share_store.get_users(merged_uid_list)

                    node_id_to_rsp_dict = defaultdict(shine_pb2.RspToUsers)

                    for row in rsp.rows:
                        node_id_to_row_dict = dict()

                        uid_list = row.uids
                        if set((constants.CONNS_AUTHED, constants.CONNS_ALL, constants.CONNS_UNAUTHED)) & set(uid_list):
                            # 给所有的topic都发一遍就好
                            node_id_list = self.share_store.get_nodes()
                            for node_id in node_id_list:
                                # 所有proc都要收到这个消息并进行处理
                                node_id_to_rsp_dict[node_id].rows.extend([row])

                            # 直接跳到下一个row
                            continue

                        for uid in uid_list:
                            if uid in row.exclude:
                                # 如果是在exclude列表里面的话，就可以不用处理
                                continue

                            node_id = node_id_to_uid_dict.get(uid)
                            if node_id is None:
                                continue

                            if node_id not in node_id_to_row_dict:
                                new_row = shine_pb2.RspToUsers.Row()
                                new_row.userdata = row.userdata
                                new_row.buf = row.buf
                                if row.exclude:
                                    new_row.exclude.extend(row.exclude)
                                node_id_to_row_dict[node_id] = new_row
                            else:
                                new_row = node_id_to_row_dict[node_id]

                            new_row.uids.append(uid)

                        for node_id, new_row in node_id_to_row_dict.items():
                            # 得用extend才行
                            node_id_to_rsp_dict[node_id].rows.extend([new_row])

                    # 消息已经搞定了，现在就是发送了
                    for node_id, rsp in node_id_to_rsp_dict.items():
                        rsp_task = shine_pb2.Task()
                        rsp_task.cmd = task.cmd
                        rsp_task.node_id = node_id
                        rsp_task.body = rsp.SerializeToString()

                        self.to_send_queue.put((node_id, rsp_task))
            elif task.cmd == constants.CMD_CLOSE_USERS:
                if not self.share_store:
                    # 直接转发就好
                    self.to_send_queue.put((task.node_id, data))
                else:
                    node_id_to_rsp_dict = defaultdict(shine_pb2.CloseUsers)

                    rsp = shine_pb2.CloseUsers()
                    rsp.ParseFromString(task.body)

                    merged_uid_list = list(rsp.uids)

                    if set((constants.CONNS_AUTHED, constants.CONNS_ALL, constants.CONNS_UNAUTHED)) & set(merged_uid_list):
                        # 给所有的topic都发一遍就好
                        node_id_list = self.share_store.get_nodes()
                        for node_id in node_id_list:
                            # 所有proc都要收到这个消息并进行处理
                            node_id_to_rsp_dict[node_id] = rsp
                    else:
                        node_id_to_uid_dict = self.share_store.get_users(merged_uid_list)

                        for uid in merged_uid_list:
                            if uid in rsp.exclude:
                                # 在排除列表里
                                continue

                            node_id = node_id_to_uid_dict.get(uid)
                            if node_id is None:
                                continue

                            if node_id not in node_id_to_rsp_dict:
                                new_rsp = shine_pb2.CloseUsers()
                                new_rsp.userdata = rsp.userdata
                                if rsp.exclude:
                                    new_rsp.exclude.extend(rsp.exclude)
                                node_id_to_rsp_dict[node_id] = new_rsp
                            else:
                                new_rsp = node_id_to_rsp_dict[node_id]

                            new_rsp.uids.append(uid)

                    # 消息已经搞定了，现在就是发送了
                    for node_id, rsp in node_id_to_rsp_dict.items():
                        rsp_task = shine_pb2.Task()
                        rsp_task.cmd = task.cmd
                        rsp_task.node_id = node_id
                        rsp_task.body = rsp.SerializeToString()

                        self.to_send_queue.put((node_id, rsp_task))

    def _handle_input_forever(self):
        """
        一直在收消息
        :return:
        """

        while 1:
            data = self.input_server.recv()

            self.to_deal_queue.put(data)

    def _handle_output_forever(self):
        """
        :return:
        """
        while 1:
            topic, data = self.to_send_queue.get()
            if isinstance(data, shine_pb2.Task):
                data = data.SerializeToString()

            self.output_server.send_multipart(
                (topic, data)
            )

    def _serve_forever(self):
        """
        保持运行
        :return:
        """
        job_list = []
        for action in [self._handle_input_forever, self._handle_task_forever, self._handle_output_forever]:
            job = gevent.spawn(action)
            job_list.append(job)

        for job in job_list:
            job.join()

    def _start_input_server(self, address):
        """
        zmq的pull server
        每个worker绑定的地址都要不一样
        """
        ctx = zmq.Context()
        self.input_server = ctx.socket(zmq.PULL)
        self.input_server.bind(address)

    def _start_output_server(self, address):
        """
        zmq的output server
        每个worker绑定的地址都要不一样
        """
        ctx = zmq.Context()
        self.output_server = ctx.socket(zmq.PUB)
        self.output_server.bind(address)

    def _worker_run(self, index):
        """
        在worker里面执行的
        :return:
        """
        setproctitle.setproctitle(self._make_proc_name('forwarder:worker:%s' % index))
        self._handle_child_proc_signals()

        self._start_input_server(self.config['FORWARDER_INPUT_ADDRESS_LIST'][index])
        self._start_output_server(self.config['FORWARDER_OUTPUT_ADDRESS_LIST'][index])

        try:
            self._serve_forever()
        except KeyboardInterrupt:
            pass
        except:
            logger.error('exc occur.', exc_info=True)

    def _handle_parent_proc_signals(self):
        def exit_handler(signum, frame):
            self.proc_mgr.enable = False

            # 如果是终端直接CTRL-C，子进程自然会在父进程之后收到INT信号，不需要再写代码发送
            # 如果直接kill -INT $parent_pid，子进程不会自动收到INT
            # 所以这里可能会导致重复发送的问题，重复发送会导致一些子进程异常，所以在子进程内部有做重复处理判断。
            self.proc_mgr.terminate_all()

        # INT, QUIT, TERM为强制结束
        signal.signal(signal.SIGINT, exit_handler)
        signal.signal(signal.SIGQUIT, exit_handler)
        signal.signal(signal.SIGTERM, exit_handler)

    def _handle_child_proc_signals(self):
        def exit_handler(signum, frame):
            # 防止重复处理KeyboardInterrupt，导致抛出异常
            if self.proc_mgr.enable:
                self.proc_mgr.enable = False
                raise KeyboardInterrupt

        # 强制结束，抛出异常终止程序进行
        signal.signal(signal.SIGINT, exit_handler)
        signal.signal(signal.SIGQUIT, exit_handler)
        signal.signal(signal.SIGTERM, exit_handler)

    def __repr__(self):
        return '<%s name: %s>' % (
            type(self).__name__, self.name
        )



# -*- coding: utf-8 -*-

import functools

from ...share.log import logger


def safe_call(func, *args, **kwargs):
    """
    安全调用
    """
    try:
        return func(*args, **kwargs)
    except Exception, e:
        logger.error('exc occur. e: %s, func: %s', e, func, exc_info=True)
        # 调用方可以通过 isinstance(e, BaseException) 来判断是否发生了异常
        return e


def safe_func(func):
    """
    把函数变为安全的
    """
    @functools.wraps(func)
    def func_wrapper(*args, **kwargs):
        return safe_call(func, *args, **kwargs)
    return func_wrapper



# -*- coding: utf-8 -*-

from ..share.shine_pb2 import Task, RspToUsers, CloseUsers
from ..share import constants
from ..share.utils import import_module_or_string


class Trigger(object):

    box_class = None

    forwarder_input_client = None

    def __init__(self, box_class=None, forwarder_input_address_list=None, forwarder_input_client=None, use_gevent=False):
        """
        :param box_class: module or string
        :param forwarder_input_address_list:
        :param forwarder_input_client:
        :param use_gevent:
        :return:
        """

        assert not (forwarder_input_address_list is forwarder_input_client is None)

        self.box_class = import_module_or_string(box_class or constants.DEFAULT_CONFIG['BOX_CLASS'])

        if forwarder_input_address_list is not None:
            if use_gevent:
                import zmq.green as zmq  # for gevent
            else:
                import zmq

            ctx = zmq.Context()
            self.forwarder_input_client = ctx.socket(zmq.PUSH)
            for address in forwarder_input_address_list:
                self.forwarder_input_client.connect(address)
        else:
            self.forwarder_input_client = forwarder_input_client

    def write_to_users(self, data_list):
        """
        格式为
        [(uids, box), (uids, box, userdata), (uids, box, userdata, exclude) ...]
        :param data_list: userdata可不传，默认为0，conn.userdata & userdata == userdata; exclude 代表排除的uid列表
        :return:
        """

        msg = RspToUsers()

        for data_tuple in data_list:
            if len(data_tuple) == 2:
                uids, data = data_tuple
                userdata = None
                exclude = None
            elif len(data_tuple) == 3:
                uids, data, userdata = data_tuple
                exclude = None
            else:
                uids, data, userdata, exclude = data_tuple

            if isinstance(data, self.box_class):
                data = data.pack()
            elif isinstance(data, dict):
                data = self.box_class(data).pack()

            row = msg.rows.add()
            row.buf = data
            row.userdata = userdata or 0
            row.uids.extend(uids)
            if exclude:
                row.exclude.extend(exclude)

        task = Task()
        task.cmd = constants.CMD_WRITE_TO_USERS
        task.body = msg.SerializeToString()

        return self._send_task(task)

    def close_users(self, uids, userdata=None, exclude=None):
        msg = CloseUsers()
        msg.uids.extend(uids)
        msg.userdata = userdata or 0
        if exclude:
            msg.exclude.extend(exclude)

        task = Task()
        task.cmd = constants.CMD_CLOSE_USERS
        task.body = msg.SerializeToString()

        return self._send_task(task)

    def write_to_worker(self, data, node_id=None):
        """
        透传到worker进行处理
        还是得把node_id传过来，告诉forwarder，是要哪个node_id进行处理
        如果node_id不传，那么forwarder就会随机选择一个node_id来处理
        """

        task = Task()
        if node_id is not None:
            task.node_id = node_id
        task.cmd = constants.CMD_WRITE_TO_WORKER

        if isinstance(data, self.box_class):
            # 打包
            data = data.pack()
        elif isinstance(data, dict):
            data = self.box_class(data).pack()

        task.body = data

        return self._send_task(task)

    def write_to_client(self, req_task, data):
        """
        写回
        :param data: 可以是dict也可以是box
        :return:
        """

        if isinstance(data, self.box_class):
            data = data.pack()
        elif isinstance(data, dict):
            req_box = self.box_class()
            if req_box.unpack(req_task.body) <= 0:
                # 解析失败了
                return False
            data = req_box.map(data).pack()

        task = Task()
        # 就可以直接通过node_id和client_id来进行识别了
        task.client_id = req_task.client_id
        task.node_id = req_task.node_id
        task.cmd = constants.CMD_WRITE_TO_CLIENT
        task.body = data

        return self._send_task(task)

    def close_client(self, req_task):
        task = Task()
        task.client_id = req_task.client_id
        task.node_id = req_task.node_id
        task.cmd = constants.CMD_CLOSE_CLIENT

        return self._send_task(task)

    def login_client(self, req_task, uid, userdata=None):

        task = Task()
        task.client_id = req_task.client_id
        task.node_id = req_task.node_id
        task.cmd = constants.CMD_LOGIN_CLIENT
        task.uid = uid
        task.userdata = userdata or 0

        return self._send_task(task)

    def logout_client(self, req_task):
        task = Task()
        task.client_id = req_task.client_id
        task.node_id = req_task.node_id
        task.cmd = constants.CMD_LOGOUT_CLIENT

        return self._send_task(task)

    def _send_task(self, task):
        """
        发送
        :return:
        """

        return self.forwarder_input_client.send(task.SerializeToString())



# -*- coding: utf-8 -*-


from gevent.server import StreamServer
from netkit.stream import Stream

from mixins import AppEventsMixin
from connection import Connection


class Server(AppEventsMixin):
    server_class = StreamServer
    connection_class = Connection
    stream_class = Stream

    box_class = None

    # 连接最长不活跃时间
    conn_timeout = None

    backlog = None
    server = None

    def __init__(self, box_class, backlog, conn_timeout=None):
        AppEventsMixin.__init__(self)
        self.box_class = box_class
        self.backlog = backlog
        self.conn_timeout = conn_timeout

    def _handle_stream(self, sock, address):
        self.connection_class(
            self, self.stream_class(sock, use_gevent=True), address
        ).handle()

    def _prepare_server(self, address):
        import socket
        # 只有这样，才能保证在主进程里面，不会启动accept
        listener = self.server_class.get_listener(address, backlog=self.backlog, family=socket.AF_INET)
        self.server = self.server_class(listener, handle=self._handle_stream)

    def _serve_forever(self):
        self.server.serve_forever()




# -*- coding: utf-8 -*-

"""
负责清理过期的数据
"""

from ..share import constants
from ..share.config import Config


class Cleaner(object):

    def __init__(self):
        self.config = Config(defaults=constants.DEFAULT_CONFIG)

    def clean_redis(self):
        if not self.config['REDIS_URL']:
            # 没有配置redis的话，就直接返回就好
            return

        import redis
        rds = redis.from_url(self.config['REDIS_URL'])
        user_key_prefix = self.config['REDIS_KEY_SHARE_PREFIX'] + self.config['REDIS_USER_KEY_PREFIX']
        nodes_key = self.config['REDIS_KEY_SHARE_PREFIX'] + self.config['REDIS_NODES_KEY']

        rds.delete(nodes_key)

        for user_key in rds.keys('%s*' % user_key_prefix):
            rds.delete(user_key)



# -*- coding: utf-8 -*-

NAME = 'shine'

# 系统返回码
RET_INVALID_CMD = -10000
RET_INTERNAL = -10001

# 命令字
CMD_CLIENT_REQ              = 10  # 透传client请求
CMD_CLIENT_CREATED          = 15  # 客户端连接建立
CMD_CLIENT_CLOSED           = 20  # 客户端连接被关闭

CMD_WRITE_TO_WORKER         = 100 # trigger触发请求

CMD_WORKER_ASK_FOR_JOB      = 210 # 请求任务

CMD_WRITE_TO_CLIENT         = 220 # 回应
CMD_WRITE_TO_USERS          = 230 # 主动下发
CMD_CLOSE_CLIENT            = 240 # 关闭客户端(client_id为判断)
CMD_CLOSE_USERS             = 250 # 关闭多个客户端
CMD_LOGIN_CLIENT            = 260 # 登录用户
CMD_LOGOUT_CLIENT           = 270 # 登出用户


# write_to_users / close_users 时的特殊连接集合
CONNS_AUTHED                = -1 # 所有已登录连接
CONNS_ALL                   = -2 # 所有连接
CONNS_UNAUTHED              = -3 # 所有未登录连接


# worker的env
WORKER_ENV_KEY = 'SHINE_WORKER'


# 默认config参数
DEFAULT_CONFIG = {
    # 进程名
    'NAME': NAME,

    # gateway, worker, forwarder 需要
    'DEBUG': False,

    # box class
    'BOX_CLASS': 'netkit.box.Box',

    # gateway 需要
    'GATEWAY_OUTER_HOST': None,
    # gateway 需要
    'GATEWAY_OUTER_PORT': None,
    # gateway 需要
    'GATEWAY_BACKLOG': 256,

    # gateway 需要
    'GATEWAY_CLIENT_HEARTBEAT_CMD': None,  # 客户端心跳的命令字，redis通过这个cmd来进行续期

    # gateway 需要
    'GATEWAY_CLIENT_TIMEOUT': None,  # 客户端连接最长不活跃时间，超过会被关闭。可以设置长一点，因为正常的关闭连接可以被检测到的。

    # gateway, worker 需要
    'GATEWAY_INNER_ADDRESS_LIST': None,

    # worker, forwarder 需要
    'FORWARDER_INPUT_ADDRESS_LIST': None,
    # gateway, forwarder 需要
    'FORWARDER_OUTPUT_ADDRESS_LIST': None,

    # worker 需要
    'WORKER_SPAWN_COUNT': 1,  # 启动多少worker

    # worker 需要
    'WORKER_RSP_ONCE': True,  # worker只能回应一次

    # worker 需要
    'WORKER_CONN_TIMEOUT': 3,  # 连接超时(秒)，比如recv一次的超时

    # worker 需要
    'WORKER_WORK_TIMEOUT': None,  # 处理超时

    # worker 需要
    'WORKER_STOP_TIMEOUT': None,  # 停止进程超时

    # gateway, forwarder 需要
    'REDIS_URL': None,  # 用来存储用户ID->node_id的映射 以及 node_id的集合
    # gateway, forwarder 需要
    'REDIS_KEY_SHARE_PREFIX': NAME + ':',  # 存储的key的统一前缀
    # gateway, forwarder 需要
    'REDIS_USER_KEY_PREFIX': 'user:',  # 存储的user键前缀
    # gateway 需要
    'REDIS_USER_MAX_AGE': None,  # user最长存储的秒数，因为有可能有些用户的数据没有正常清空
    # gateway, forwarder 需要
    'REDIS_NODES_KEY': 'nodes',  # 存储node_id的集合
}
